# -*- coding: utf-8 -*-
"""
    Toolbox for SCS.

    Various utilities function to quickly process data measured at the SCS
    instruments.

    Copyright (2019) SCS Team.
"""

import logging

import numpy as np
import xarray as xr

import extra_data as ed

from .misc.bunch_pattern import extractBunchPattern
from .constants import mnemonics as _mnemonics_ld
from .util.exceptions import ToolBoxValueError, ToolBoxPathError
from .util.data_access import find_run_dir

log = logging.getLogger(__name__)


def load(fields, runNB, proposalNB,
         subFolder='raw',
         display=False,
         validate=False,
         subset=ed.by_index[:],
         rois={},
         useBPTable=True):
    """
    Load a run and extract the data. Output is an xarray with aligned
    trainIds

    Parameters
    ----------

    fields: list of strings, list of dictionaries
        list of mnemonic strings to load specific data such as "fastccd",
        "SCS_XGM", or dictionnaries defining a custom mnemonic such as
            {"extra":
                {'SCS_CDIFFT_MAG/SUPPLY/CURRENT',
                 'actual_current.value', None}}
    runNB: (str, int)
        run number as integer
    proposalNB: (str, int)
        of the proposal number e.g. 'p002252' or 2252
    subFolder: (str)
        sub-folder from which to load the data. Use 'raw' for raw data
        or 'proc' for processed data.
    display: (bool)
        whether to show the run.info or not
    validate: (bool)
        whether to run extra-data-validate or not
    subset:
        a subset of train that can be load with by_index[:5] for the first 5
        trains
    rois: dictionary
        a dictionnary of mnemonics with a list of rois definition and
        the desired names, for example:
            {'fastccd':
                {'ref':
                    {'roi': by_index[730:890, 535:720],
                     'dim': ['ref_x', 'ref_y']},
                 'sam':
                    {'roi':by_index[1050:1210, 535:720],
                     'dim': ['sam_x', 'sam_y']}}}
    useBPTable: boolean
        If True, uses the raw bunch pattern table to extract sase pulse number
        and indices in the trains. If false, load the data from BUNCH_DECODER
        middle layer device.

    Returns
    -------
    res: xarray.DataArray
        an xarray DataSet with aligned trainIds
    """
    try:
        runFolder = find_run_dir(proposalNB, runNB)
    except ToolBoxPathError as err:
        log.error(f"{err.message}")
        raise

    run = ed.RunDirectory(runFolder).select_trains(subset)

    if validate:
        # get_ipython().system('extra-data-validate ' + runFolder)
        pass
    if display:
        print('Loading data from {}'.format(runFolder))
        run.info()

    keys = []
    vals = []

    # load pulse pattern info
    if useBPTable:
        bp_mnemo = _mnemonics_ld['bunchPatternTable']
        if bp_mnemo['source'] not in run.all_sources:
            print('Source {} not found in run. Skipping!'.format(
                                _mnemonics_ld['bunchPatternTable']['source']))
        else:
            bp_table = run.get_array(bp_mnemo['source'], bp_mnemo['key'],
                                     extra_dims=bp_mnemo['dim'])
            sase1, npulses_sase1, dummy = extractBunchPattern(
                                              bp_table, 'sase1')
            sase3, npulses_sase3, dummy = extractBunchPattern(
                                              bp_table, 'sase3')
            keys += ["sase1", "npulses_sase1", "sase3", "npulses_sase3"]
            vals += [sase1, npulses_sase1, sase3, npulses_sase3]
    else:
        fields += ["sase1", "sase3", "npulses_sase3", "npulses_sase1"]

    for f in fields:

        if type(f) == dict:
            # extracting mnemomic defined on the spot
            if len(f.keys()) > 1:
                print('Loading only one "on-the-spot" mnemonic at a time, '
                      'skipping all others !')
            k = list(f.keys())[0]
            v = f[k]
        else:
            # extracting mnemomic from the table
            if f in _mnemonics_ld:
                v = _mnemonics_ld[f]
                k = f
            else:
                print('Unknow mnemonic "{}". Skipping!'.format(f))
                continue

        if k in keys:
            continue  # already loaded, skip

        if display:
            print('Loading {}'.format(k))

        if v['source'] not in run.all_sources:
            print('Source {} not found in run. Skipping!'.format(v['source']))
            continue

        if k not in rois:
            # no ROIs selection, we read everything
            vals.append(run.get_array(v['source'], v['key'],
                                      extra_dims=v['dim']))
            keys.append(k)
        else:
            # ROIs selection, for each ROI we select a region of the data and
            # save it with new name and dimensions
            for nk, nv in rois[k].items():
                vals.append(run.get_array(v['source'], v['key'],
                                          extra_dims=nv['dim'],
                                          roi=nv['roi']))
                keys.append(nk)

    aligned_vals = xr.align(*vals, join='inner')
    result = dict(zip(keys, aligned_vals))
    result = xr.Dataset(result)
    result.attrs['run'] = run
    result.attrs['runFolder'] = runFolder
    return result


def load_run(proposal, run, **kwargs):
    """
    Get run in given proposal

    Wraps the extra_data open_run routine, out of convenience for the toolbox
    user. More information can be found in the karabo_data documentation.

    Parameters
    ----------
    proposal: str, int
        Proposal number
    run: str, int
        Run number

    **kwargs
    --------
    data: str
        default -> 'raw'
    include: str
        default -> '*'

    Returns
    -------
    run : extra_data.DataCollection
        DataCollection object containing information about the specified
        run. Data can be loaded using built-in class methods.
    """
    return ed.open_run(proposal, run, **kwargs)


def run_by_path(path):
    """
    Return specified run

    Wraps the extra_data RunDirectory routine, to ease its use for the
    scs-toolbox user.

    Parameters
    ----------
    path: str
        path to the run directory

    Returns
    -------
    run : extra_data.DataCollection
        DataCollection object containing information about the specified
        run. Data can be loaded using built-in class methods.
    """
    return ed.RunDirectory(path)


def concatenateRuns(runs):
    """ Sorts and concatenate a list of runs with identical data variables
        along the trainId dimension.

        Input:
            runs: (list) the xarray Datasets to concatenate
        Output:
            a concatenated xarray Dataset
    """
    firstTid = {i: int(run.trainId[0].values) for i, run in enumerate(runs)}
    orderedDict = dict(sorted(firstTid.items(), key=lambda t: t[1]))
    orderedRuns = [runs[i] for i in orderedDict]
    keys = orderedRuns[0].keys()
    for run in orderedRuns[1:]:
        if run.keys() != keys:
            print('data fields between different runs are not identical. '
                  'Cannot combine runs.')
            return

    result = xr.concat(orderedRuns, dim='trainId')
    for k in orderedRuns[0].attrs.keys():
        result.attrs[k] = [run.attrs[k] for run in orderedRuns]
    return result


def get_array(run, mnemonic_key=None, stepsize=None):
    """
    Loads the required 1D-data and rounds its values to integer multiples of
    stepsize for consistent grouping (except for stepsize=None).
    Returns a dummy array if mnemonic is set to None.

    Parameters
    ----------
    run: karabo_data.DataCollection
        path to the run directory
    mnemonic_key: str
        Identifier of a single item in the mnemonic collection. None creates a
        dummy file to average over all trains in the run
    stepsize : float
        nominal stepsize of the array data - values will be rounded to integer
        multiples of this value

    Returns
    -------
    data : xarray.DataArray
        xarray DataArray containing rounded array values using the trainId as
        coordinate.

    Raises
    ------
    ToolBoxValueError: Exception
        Toolbox specific exception, indicating a non-valid mnemonic entry

    Example
    -------
    >>> import toolbox_scs as tb
    >>> run = tb.load_run(2212, 235)
    >>> mnemonic = 'PP800_PhaseShifter'
    >>> data_PhaseShifter = tb.get_array(run, mnemonic, 0.5)
    """

    try:
        if mnemonic_key is None:
            data = xr.DataArray(
                        np.ones(len(run.train_ids), dtype=np.int16),
                        dims=['trainId'], coords={'trainId': run.train_ids})
        elif mnemonic_key in _mnemonics_ld:
            mnem = _mnemonics_ld[mnemonic_key]
            data = run.get_array(*mnem.values())
        else:
            raise ToolBoxValueError("Invalid mnemonic", mnemonic_key)

        if stepsize is not None:
            data = stepsize * np.round(data / stepsize)
        data.name = 'data'
        log.debug(f"Got data for {mnemonic_key}")
    except ToolBoxValueError as err:
        log.error(f"{err.message}")
        raise

    return data
